---
sidebar_position: 1
description: "在游戏开发中享受不用关注 Draw Call 的快乐。"
---

# 新 UI 渲染批次合并指南

在官方文档的进阶主题中，有一个 [UI 渲染批次合并指南](https://docs.cocos.com/creator/2.4/manual/zh/advanced-topics/ui-auto-batch.html)，而由于服务包的 **多纹理渲染**、**重构动态图集** 等新特性的出现，有一些内容可能已经过期。

如果你未阅读过官方的指南，可以先看一看里面的知识科普部分，而我们将直接从如何实践开始！

---
## 什么是多纹理渲染？

在以前的认识里，我们知道相邻的节点使用不同的纹理（Texture）会导致不能合并批次。

其根本原因是纹理是使用 uniform 变量传给着色器的，而需要合并批次的话不允许每次渲染都拥有不同的 uniform 变量值。

服务包实现的合批方法是先设置好多个 uniform 变量，比如将 8 张纹理写入到 "texture1" "texture2" "texture3"... 的 8 个 uniform 变量中，然后编写一个着色器判断应该在渲染时使用哪个 uniform 变量。

这样的话如果所有渲染都只用这 8 张纹理的话，就都能合并为 1 个批次。

这种做法要求设备需要支持采样多个纹理，而在现代绝大多数设备中这不是问题，至少都支持采样 8 张纹理。

当然除了这种方法，还有另外的一些进行多纹理合批的方法，例如 "Texture Array"、"Bindless"，但都有实用性与兼容性的问题。

:::info 提示

以这种方式实现的多纹理渲染并不是没有弊端的：

因为会多传递一个顶点属性，并且需要在着色器中去判断该使用哪个纹理，这可能也会造成性能下降，毕竟**合并批次并不一定会提升性能**。

所以我们建议在多个档次设备中实际测试项目是否使用多纹理渲染的性能差距。

如果你还有所担心，就像我们在实现这个功能之前一样，可以看看这些：

**PixiJS 引擎在 2016 年发布的 v4 版本就已经正式实装了多纹理渲染机制！**[资料出处](https://medium.com/goodboy-digital/gpu-multi-texture-sprite-batching-21c90ae8f89b)。

**Phaser 引擎在 v4 与 v3 版本都实装了多纹理渲染机制（2019 - 2020年）**[资料出处](https://www.patreon.com/posts/39665256)。

:::

---
## 为你的项目启用动态合图

在项目之前的开发中，我们可能会关闭动态图集，更倾向于靠静态图集或者自动图集达到降低 Draw Call 的目的。

这种选择除了与动态合图存在会占用更多内存的缺点有关之外，更多的是在引擎原来的动态合图中，并不能复用图集的废弃区域，这个重要特性的缺失只能靠在切换场景（Scene）后重置所有图集来解决。

但对于大部分项目来说，这种治标不治本的机制基本等于没有解决这个问题。

而现在，**服务包几乎重构了整个动态合图系统，解决了以前存在的大部分问题**，你可以启用它了。

:::note 提示

开启动态图集常见的反对意见是：

在部分小游戏平台里，启用动态图集会有保留 Image 对象所占用的内存空间很大的问题。

我们建议：

- 请实际测试是否启用动态图集的内存占用差距。
- 有没有一种可能，只是说可能，出现不能接受的内存占用大小是因为你的项目根本就不做任何资源的释放呢？

:::

---
## 充分利用动态合图

启用动态合图后，我们有几个小提示能让你发挥出动态合图的潜力：

### 放宽能参与合图的纹理尺寸限制

你可以适当地放宽能参与合图的纹理尺寸的限制（`cc.dynamicAtlasManager.maxFrameSize`）。

推荐设置为 `512`、 `1024` 甚至 `2048`。

因为**动态图集会自动进行多纹理合批，你可以放心地使用多达 8 张图集而不用担心交叉渲染导致的打断批次！**

:::tip 

服务包会自动将图集最大数量调整至（设备能同时采样纹理数 - Char 缓存模式自动合批图集数），这个值默认为 `7`。

:::

### 无需管理动态图集，只需要释放资源

**动态合图支持复用废弃空间，释放纹理的同时会释放其在动态图集使用的空间。**

所以一般不需要去关心动态图集，只需要做好资源释放，就不会发生动态图集被用完的情况。

### 更加细致地优化图集的使用效率

除了引擎自带的 `packable` 属性可以控制纹理是否会参与动态合图之外。

服务包还额外支持了**控制组件是否默认参与动态合图，也可以控制单个组件是否参与动态合图。**

有关动态合图的设置可前往 [动态合图](TODO) 的文档了解详情。

在上面我们推荐可以将纹理尺寸限制放宽到 `2048`，这听起来貌似有点离谱，但只要规划好项目的资源确实可行，比如：

- 将优化程度有限但尺寸具大的纹理禁止参与动态合图
- 分模块存放资源，将冷门界面（如活动）的纹理禁止参与动态合图，或尽早释放掉
- 在资源已经一团糟的项目中，可以直接禁止渲染组件参与动态合图

---
## Label 不再是合批噩梦

在项目之前的开发中，我们可能会使用字体图集、调整节点顺序，甚至修改渲染流程来解决 Label 的性能问题。

但是引擎不是提供了 Bitmap 和 Char 两种缓存模式吗？

是的，但是这两个解决方案与旧的动态合图一样在生产环境中使用太过 "玩具" 了：

- Bitmap 缓存模式的缺点：字体纹理会打入动态图集，而动态图集无法复用，随着游戏的进行，图集用完则直接失去作用。

- Char 缓存模式的缺点：还是无法复用，并且只有一张图集，用完则直接无法渲染，试问谁能接受游戏可能跑着跑着字就全部消失了的情况。

所以，**服务包几乎重构了 Char 缓存模式，除了解决不能复用的问题之外，由于支持了多纹理渲染，所以既可以与动态图集一起合批，还拥有高达最多 8 张字体图集可以使用！**

接下来我们有几个小提示能让你知道如何选择合适的缓存模式：

### 委以重任一把梭 —— Char

遇事不决，先选 Char 缓存模式。

虽然 Char 模式也有一些缺点，但能够应付大多数的场景，由于它既能与动态图集一起合批，还是是按字符进行复用的，所以相比 Bitmap 模式它有着更高的性能优势。

而且不用担心这 8 张字符图集会被用完，内部会用引用计数自动释放废弃字符所占用的空间，可以试着想一下你的游戏最多在同一帧会显示几个字符。

但 Char 缓存模式不适合下面的场景：

- 无法显示带有像 emoji 的字素簇的字符串，这种字符串现在不能被完美地分割成单个字符，所以 Char 缓存模式也就不能正常显示了。
- 接上条，像聊天消息、输入框这类不可控的内容文本都不建议用 Char 缓存模式。
- Char 缓存模式不支持一些字体样式，可以在官方文档中了解详情。
- 巨大的字体大小（比如几百的）可能会瞬间占满整张字符图集，字符图集虽然有 8 张但也不能这么霍霍。

### 最后的合批守护者 —— Bitmap

即使不能选择 Char 缓存模式，Bitmap 缓存模式也能成为批次的最后一道防线。

在解决了动态图集的复用问题后，Bitmap 缓存模式的纹理也会使用引用计数自动释放，并且不会有 Char 缓存模式无法显示字素簇的问题。

但当然，Bitmap 缓存模式也不是万能的，如果遇到了下面这种情况就再考虑调整节点顺序等方法来解决吧：

- 巨大的字体大小也会瞬间占满整张动态图集
- 在非常大量的 Label 需频繁改变文本的情况下，请使用性能分析工具检查动态图集的性能消耗，避免合批的弊大于利。

### 关于字符图集与动态图集

常见的误区：

#### 1.Char 缓存模式所使用的字符图集与动态图集是一样的东西，只是叫法不同

由于多种原因，没有设计让 Char 缓存模式直接使用动态图集，这当然可以做到，但是权衡之下选择了分开实现，在使用指南中有详细解释。

多纹理材质只有 8 个纹理插槽，所以默认情况下动态图集最大数量为 7，字符图集自动多纹理合批的数量为 1。（注意不是最大数量）

你可以自己调整这个分配值，而服务包使用这个默认值有以下几点原因：

- 引擎原本就只有 1 张 Char 字符图集
- 大多数项目因为 Char 缓存模式加入了可复用的特性后，1 张 Char 字符图集是足够的

如果你直觉认为 1 张太少，我们建议你合理搭配使用 Bitmap 和 Char 两种缓存模式后，实际测试发现不够再调整该值。

**请注意上面的 1 并非是字符图集的最大数量，而是字符图集进行自动多纹理合批的最大数量！**

所以这并不意味着字符图集只能有 1 张，而是最大能有 8 张，并且这个数量不能调整。

如果你还是对为什么会出现一个 “字符图集进行自动多纹理合批的最大数量” 而一头雾水，建议详细阅读使用指南。

#### 2.Char 缓存模式能在图集满了的情况下依然能正常显示

现在如果字符图集满了，依旧不能正常显示，这样的原因有以下几点：

- 我们认为 8 张数量已经够多了，8 张都用完的情况大部分是没有合理搭配使用两种缓存模式
- 8 张是多纹理渲染的上限，这意味着如果超过 8 张，1 个 Label 有 100 个字，就可能有 100 个 Draw Call

---
## 总结

以上就是新合批指南的全部内容了，稍微总结一下渲染批次合并的几个要点：

- 启用动态合图，只需要合理地释放资源即可保持动态合图的一直有效
- 能就优先用 Char 缓存模式，不能则换为 Bitmap 缓存模式，字数太多字体太大的则放弃挣扎使用 None 模式
- 不要优先考虑打图集，修改节点顺序等需要维护成本的优化方式

如果你对批次合并还有着更高的需求，可以阅读 [进阶合批指南](./advance-batcher-guide)。
